@@ -132,14 +132,13 @@ class XdebugHandler
 
             if ($this->prepareRestart()) {
                 $command = $this->getCommand();
-                $this->notify(Status::RESTARTING, $command);
                 $this->restart($command);
             }
             return;
         }
 
         if (self::RESTART_ID === $envArgs[0] && count($envArgs) === 5) {
-            // Restarting, so unset environment variable and use saved values
+            // Restarted, so unset environment variable and use saved values
             $this->notify(Status::RESTARTED);
 
             Process::setEnv($this->envAllowXdebug);
@@ -150,6 +149,8 @@ class XdebugHandler
                 self::$skipped = $envArgs[1];
             }
 
+            $this->tryEnableSignals();
+
             // Put restart settings in the environment
             $this->setEnvRestartSettings($envArgs);
             return;
@@ -256,8 +257,29 @@ class XdebugHandler
      */
     private function doRestart($command)
     {
-        passthru($command, $exitCode);
-        $this->notify(Status::INFO, 'Restarted process exited '.$exitCode);
+        $this->tryEnableSignals();
+        $this->notify(Status::RESTARTING, $command);
+
+        // Prefer proc_open to keep fds intact, because passthru pipes to stdout
+        if (function_exists('proc_open')) {
+            if (defined('PHP_WINDOWS_VERSION_BUILD') && PHP_VERSION_ID < 80000) {
+                $command = '"'.$command.'"';
+            }
+            $process = proc_open($command, array(), $pipes);
+            if (is_resource($process)) {
+                $exitCode = proc_close($process);
+            }
+        } else {
+            passthru($command, $exitCode);
+        }
+
+        if (!isset($exitCode)) {
+            // Unlikely that the default shell cannot be invoked
+            $this->notify(Status::ERROR, 'Unable to restart process');
+            $exitCode = -1;
+        } else {
+            $this->notify(Status::INFO, 'Restarted process exited '.$exitCode);
+        }
 
         if ($this->debug === '2') {
             $this->notify(Status::INFO, 'Temp ini saved: '.$this->tmpIni);
@@ -333,7 +355,7 @@ class XdebugHandler
 
         foreach ($iniFiles as $file) {
             // Check for inaccessible ini files
-            if (!$data = @file_get_contents($file)) {
+            if (($data = @file_get_contents($file)) === false) {
                 $error = 'Unable to read ini: '.$file;
                 return false;
             }
@@ -562,4 +584,32 @@ class XdebugHandler
 
         return true;
     }
+
+    /**
+     * Enables async signals and control interrupts in the restarted process
+     *
+     * Only available on Unix PHP 7.1+ with the pcntl extension. To replicate on
+     * Windows would require PHP 7.4+ using proc_open rather than passthru.
+     */
+    private function tryEnableSignals()
+    {
+        if (!function_exists('pcntl_async_signals') || !function_exists('pcntl_signal')) {
+            return;
+        }
+
+        pcntl_async_signals(true);
+        $message = 'Async signals enabled';
+
+        if (!self::$inRestart) {
+            // Restarting, so ignore SIGINT in parent
+            pcntl_signal(SIGINT, SIG_IGN);
+            $message .= ' (SIGINT = SIG_IGN)';
+        } elseif (is_int(pcntl_signal_get_handler(SIGINT))) {
+            // Restarted, no handler set so force default action
+            pcntl_signal(SIGINT, SIG_DFL);
+            $message .= ' (SIGINT = SIG_DFL)';
+        }
+
+        $this->notify(Status::INFO, $message);
+    }
 }
